/*
**    TP CPE Lyon
**    Copyright (C) 2014 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "opengl_drawer.hpp"

#include "../libmesh/mesh.hpp"
#include "../libspline/surface_param_drawable.hpp"
#include "../libspline/grid.hpp"
#include "../lib3d/vec3.hpp"

#include <cmath>

#include <iostream>
#include <sstream>
#include <GL/glu.h>

namespace cpe
{
  void opengl_drawer::draw(const mesh& m)
  {
      unsigned int N_tri=m.number_of_triangle();
      const int* p_connectivity=&m.get_connectivity()[0];
      const float* p_vertex=m.get_vertices()[0].pointer();
      const float* p_normal=m.get_normal()[0].pointer();
      const float* p_texture=m.get_texture()[0].pointer();
      const float* p_color=m.get_color()[0].pointer();

      bool is_texture=(m.get_texture().size()==m.get_vertices().size());
      bool is_color=(m.get_color().size()==m.get_vertices().size());

      if(m.get_normal().size()!=m.get_vertices().size())
      {
          std::stringstream msg;
          msg<<"Error, incorrect normals for this mesh "<<std::endl;
          msg<<"Number of vertices: ["<<m.get_vertices().size()<<"] / Number of normales ["<<m.get_normal().size()<<"]";
          if(m.get_normal().size()==0)
              msg<<std::endl<<"(Use mesh.compute_normal() before drawing)"<<std::endl;
          throw cpe::exception_cpe(msg.str(),MACRO_EXCEPTION_PARAMETER);
      }


      //enable client state
      glEnableClientState(GL_VERTEX_ARRAY);
      glEnableClientState(GL_NORMAL_ARRAY);
      if(is_texture)
          glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      if(is_color)
          glEnableClientState(GL_COLOR_ARRAY);

      //setup pointers
      glVertexPointer(3,GL_FLOAT,0,p_vertex);
      glNormalPointer(GL_FLOAT,0,p_normal);
      if(is_texture)
          glTexCoordPointer(2,GL_FLOAT,0,p_texture);
      if(is_color)
          glColorPointer(3,GL_FLOAT,0,p_color);

      //draw triangles
      glDrawElements(GL_TRIANGLES,3*N_tri,GL_UNSIGNED_INT,p_connectivity);

      //disable client state
      glDisableClientState(GL_VERTEX_ARRAY);
      glDisableClientState(GL_NORMAL_ARRAY);
      if(is_texture)
          glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      if(is_color)
          glDisableClientState(GL_COLOR_ARRAY);

  }

  void opengl_drawer::draw(const surface_param_drawable& surf)
  {
      const unsigned int& Nu=surf.size_u();
      const unsigned int& Nv=surf.size_v();
      unsigned int N_quad=(Nu-1)*(Nv-1);

      const float* p=surf.pointer_draw_vertices();
      const float* n=surf.pointer_draw_normal();
      const unsigned int* c=surf.pointer_draw_connectivity();

      //enable client state
      glEnableClientState(GL_VERTEX_ARRAY);
      glEnableClientState(GL_NORMAL_ARRAY);

      //setup pointers
      glVertexPointer(3,GL_FLOAT,0,p);
      glNormalPointer(GL_FLOAT,0,n);

      //draw triangles
      glDrawElements(GL_QUADS,4*N_quad,GL_UNSIGNED_INT,c);

      //disable client state
      glDisableClientState(GL_VERTEX_ARRAY);
      glDisableClientState(GL_NORMAL_ARRAY);

  }

  void opengl_drawer::draw(const grid& g)
  {
      const unsigned int& Nu=g.size_u();
      const unsigned int& Nv=g.size_v();

      glBegin(GL_LINES);
      for(unsigned int ku=0;ku<Nu-1;++ku)
      {
          for(unsigned int kv=0;kv<Nv;++kv)
          {
              const vec3& x0=g(ku,kv);
              const vec3& x1=g(ku+1,kv);
              glVertex3fv(x0.pointer());
              glVertex3fv(x1.pointer());
          }
      }
      for(unsigned int ku=0;ku<Nu;++ku)
      {
          for(unsigned int kv=0;kv<Nv-1;++kv)
          {
              const vec3& x0=g(ku,kv);
              const vec3& x1=g(ku,kv+1);
              glVertex3fv(x0.pointer());
              glVertex3fv(x1.pointer());
          }
      }
      glEnd();
  }

  void opengl_drawer::draw_curve(const std::vector<std::vector<cpe::vec3> >& curves)
  {
      glBegin(GL_LINES);
      const unsigned int& N_curve=curves.size();
      for(unsigned int k=0;k<N_curve;++k)
      {
          const std::vector<cpe::vec3>& current_curve=curves[k];
          const unsigned int& N_sample=current_curve.size();
          for(unsigned int k2=0;k2<N_sample-1;++k2)
          {
              vec3 x0=current_curve[k2];
              vec3 x1=current_curve[k2+1];

              glVertex3fv(x0.pointer());
              glVertex3fv(x1.pointer());
          }
      }
      glEnd();
  }


  void opengl_drawer::draw_curve(const std::vector<cpe::vec3>& curve)
  {
      glBegin(GL_LINES);
      const unsigned int& N_sample=curve.size();
      for(unsigned int k=0;k<N_sample-1;++k)
      {
          vec3 x0=curve[k];
          vec3 x1=curve[k+1];

          glVertex3fv(x0.pointer());
          glVertex3fv(x1.pointer());
      }
      glEnd();
  }

  void opengl_drawer::draw_sphere(const grid& g,const mesh& sphere)
  {
      const unsigned int& Nu=g.size_u();
      const unsigned int& Nv=g.size_v();
      for(unsigned int ku=0;ku<Nu;++ku)
      {
          for(unsigned int kv=0;kv<Nv;++kv)
          {
              const vec3& t=g(ku,kv);

              //translation rapide en OpenGL par Push/Pop-Matrix
              glPushMatrix();
              glTranslated(t.x(),t.y(),t.z());
              glColor3d(1,0,0);

              //affichage de la sphere
              opengl_drawer::draw(sphere);
              glPopMatrix();
          }
      }
  }


}
