/*
**    TP CPE Lyon
**    Copyright (C) 2014 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/





#include "scene.hpp"

#include <GL/glew.h>
#include <QString>
#include <QFont>
#include <cmath>
#include <iostream>
#include <fstream>

#include "../libinterface/glwidget.hpp"
#include "../libinterface/opengl_drawer.hpp"
#include "../libinterface/string_helper.hpp"
#include "../libinterface/intersection_helper.hpp"

#include "../lib3d/matrix4.hpp"
#include "../lib3d/matrix3.hpp"

#include "../libspline/patch4.hpp"
#include "../libspline/surface_param.hpp"
#include "../libspline/surface_param_drawable.hpp"
#include "../libspline/evaluator_spline.hpp"


using namespace cpe;





//************************************************************//
//*** Ici, on charge les objets de la scène une seule fois ***//
//************************************************************//
void scene::load_model()
{

    // *********************************************** //
    // Chargement d'une sphere sous forme de maillage
    // *********************************************** //
    preload_sphere();



    // *********************************************** //
    // Chargement de la grille de controle
    // *********************************************** //

    //definition des données suivant x
    matrix4 patch_x=matrix4(0,0,0,0,
                            1,1,1,1,
                            2,2,2,2,
                            3,3,3,3);

    //definition des données suivant y
    matrix4 patch_y=matrix4(0,1,2,3,
                            0,1,2,3,
                            0,1,2,3,
                            0,1,2,3);

    //definition des données suivant z
    matrix4 patch_z=matrix4(0,1,1,0,
                            1,2,2,1,
                            1,2,2,1,
                            0,1,1,0);



    // *********************************************** //
    // Calcul du patch courant
    // *********************************************** //

    // Le foncteur d'evaluation en (x,y,z)
    evaluator_spline evaluateur_x(patch_x);
    evaluator_spline evaluateur_y(patch_y);
    evaluator_spline evaluateur_z(patch_z);


    //nombre d'échantillons suivant u et v
    unsigned int Nu=50;
    unsigned int Nv=50;

    //redimensionnement pour stocker la surface
    surf_1.resize(Nu,Nv);

    //Pour tous les echantillons suivant u
    for(unsigned int ku=0;ku<Nu;++ku)
    {
        double u=static_cast<double>(ku)/(Nu-1);

        //Pour tous les echantillons suivant v
        for(unsigned int kv=0;kv<Nv;++kv)
        {
            double v=static_cast<double>(kv)/(Nv-1);

            // ******************************** //
            // Evaluation des coordonnees       //
            // ******************************** //

            //Evaluation de la spline en (x,y,z)
            double x=evaluateur_x(u,v);
            double y=evaluateur_y(u,v);
            double z=evaluateur_z(u,v);

            //Coordonnee de mon point en ces parametres (u,v)
            surf_1(ku,kv)=vec3(x,y,z);


            // ******************************** //
            // Evaluation de la normale         //
            // ******************************** //

            //Calcul de la derivee en u, puis en v
            vec3 du=vec3(evaluateur_x.diff_u(u,v),evaluateur_y.diff_u(u,v),evaluateur_z.diff_u(u,v));
            vec3 dv=vec3(evaluateur_x.diff_v(u,v),evaluateur_y.diff_v(u,v),evaluateur_z.diff_v(u,v));


            //Calcul de la normale
            vec3 n=(du.cross(dv)).normalized();

            //Valeur de la normale en ce point
            surf_1.set_normal(ku,kv,n);

        }
    }



    //connectivitee de la surface
    surf_1.update_connectivity();


    //mise en forme de la grille de controle pour l'affichage
    grid_1.resize(4,4);
    for(unsigned int ku=0;ku<4;++ku)
        for(unsigned int kv=0;kv<4;++kv)
            grid_1(ku,kv)=vec3(patch_x(ku,kv),patch_y(ku,kv),patch_z(ku,kv));



}



//************************************************************************//
//*** Ici, on affiche ce que l'on veut.                                ***//
//*** Cette fonction est appellee un grand nombre de fois par secondes ***//
//************************************************************************//
void scene::draw_scene()
{


    glEnable(GL_LIGHTING);//commande OpenGL pour l'eclairage



    //Affichage de la surface spline
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(6,6);

    glColor3d(0.5,0.9,0.8);

    //affiche de la surface proprement dit
    if(gui_visible.get_surface_visible()==true)
        opengl_drawer::draw(surf_1);


    glDisable(GL_POLYGON_OFFSET_FILL);

    //affichage de la grille de controle
    glLineWidth(1.5);
    glDisable(GL_LIGHTING);
    glColor3d(0.5,0.5,0.8);
    if(gui_visible.get_grid_visible()==true)
        opengl_drawer::draw(grid_1);
    glEnable(GL_LIGHTING);


    //affichage des positions de la grille en tant que sphere
    glEnable(GL_LIGHTING);
    if(gui_visible.get_gridposition_visible()==true)
        opengl_drawer::draw_sphere(grid_1,sphere);
    glDisable(GL_LIGHTING);


}








//initialisation de la classe scene
scene::scene()
{
    init_gui_visibility();
}

//retourne l'interface de sauvegarde des parametres de la gui
gui_interface& scene::gui_visible_interface()
{
    return gui_visible;
}

//initialisation des parametres de visibilite de la scene
void scene::init_gui_visibility()
{
    gui_visible.set_grid_visible(true);
    gui_visible.set_gridposition_visible(true);
    gui_visible.set_surface_visible(true);
}
//prechargement de la sphere pour les sommets de la grille
void scene::preload_sphere()
{
    std::string filename="data/sphere.off";
    std::ifstream stream(filename.c_str());
    if(stream.is_open()==false || stream.good()==false)
    {
        std::cout<<"Fichier "<<filename<<" non accessible, verifiez le chemin d'execution"<<std::endl;
        abort();
    }
    stream.close();

    sphere.load_file(filename);
    sphere.scale(0.06,0.06,0.06);
    sphere.compute_normal();
}

