/*
**    TP CPE Lyon
**    Copyright (C) 2014 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/




#include "scene.hpp"

#include <GL/glew.h>
#include <QString>
#include <QFont>
#include <cmath>
#include <fstream>
#include <iostream>
#include <fstream>

#include "../libinterface/glwidget.hpp"
#include "../libinterface/opengl_drawer.hpp"
#include "../libinterface/string_helper.hpp"
#include "../libinterface/intersection_helper.hpp"

#include "../lib3d/matrix4.hpp"
#include "../libspline/patch4.hpp"
#include "../libspline/surface_param.hpp"
#include "../libspline/surface_param_drawable.hpp"
#include "../libspline/evaluator_spline.hpp"

using namespace cpe;





//************************************************************//
//*** Ici, on charge les objets de la scène une seule fois ***//
//************************************************************//
void scene::load_model()
{


    // *********************************************** //
    // Chargement d'une sphere sous forme de maillage
    // *********************************************** //
    preload_sphere();



    // *********************************************** //
    // Chargement de la grille de controle
    // *********************************************** //

    //construction d'une grille 4x4
    g0.resize(4,4);

    //construit une gille avec z=0, (x,y)=(ku,kv)
    g0.build_meshgrid();

    //initialize les valeurs z de la grille
    g0(0,0).z()=0; g0(0,1).z()=1; g0(0,2).z()=1; g0(0,3).z()=0;
    g0(1,0).z()=1; g0(1,1).z()=2; g0(1,2).z()=2; g0(1,3).z()=1;
    g0(2,0).z()=1; g0(2,1).z()=2; g0(2,2).z()=2; g0(2,3).z()=1;
    g0(3,0).z()=0; g0(3,1).z()=1; g0(3,2).z()=1; g0(3,3).z()=0;



    // calcul de la surface
    update_high_res();




}


//************************************************************//
//***Calcul de la surface spline  ***//
//************************************************************//
void scene::build_spline_surface(const unsigned int& N_samples)
{
    const unsigned int& Nu=g0.size_u();
    const unsigned int& Nv=g0.size_v();


    //retaillage de la surface au besoin: elle contiendra (Nu-3)*(Nv-3) patches
    if(surf.size()!=(Nu-3)*(Nv-3))
        surf.resize((Nu-3)*(Nv-3));


    // Pour tous les patchs en u
    for(unsigned int k_patch_u=0;k_patch_u<Nu-3;++k_patch_u)
    {
        //Pour tous les patchs en v
        for(unsigned int k_patch_v=0;k_patch_v<Nv-3;++k_patch_v)
        {
            //On recupere le patch courant (patch = 4x4 sommets)
            patch4 p=g0.get_patch(k_patch_u,k_patch_v);

            //Voici l'indice du morceau de surface dans le vecteur concatene
            unsigned int index=k_patch_u+(Nu-3)*k_patch_v;

            //On remplit la surface correspondant a ce patch
            fill_patch(index,p,N_samples);
        }
    }

}

//************************************************************//
//***Calcul de la surface spline pour un patch donne       ***//
//************************************************************//
void scene::fill_patch(unsigned int index_surface,const patch4& p,unsigned int N_samples)
{
    unsigned int Nu=N_samples;
    unsigned int Nv=N_samples;

    //redimensionnement de la surface au besoin
    if(surf[index_surface].size_u()!=Nu || surf[index_surface].size_v()!=Nv)
        surf[index_surface].resize(Nu,Nv);

    // La matrice des coordonnees en (x,y,z)
    matrix4 Mx=p.get_x();
    matrix4 My=p.get_y();
    matrix4 Mz=p.get_z();

    // Le foncteur d'evaluation en (x,y,z)
    evaluator_spline spline_x(Mx);
    evaluator_spline spline_y(My);
    evaluator_spline spline_z(Mz);

    //Pour tous les echantillons suivant u
    for(unsigned int ku=0;ku<Nu;++ku)
    {
        double u=static_cast<double>(ku)/(Nu-1);

        //Pour tous les echantillons suivant v
        for(unsigned int kv=0;kv<Nv;++kv)
        {
            double v=static_cast<double>(kv)/(Nv-1);

            // ******************************** //
            // Evaluation des coordonnees       //
            // ******************************** //

            //Evaluation de la spline en (x,y,z)
            double x=spline_x(u,v);
            double y=spline_y(u,v);
            double z=spline_z(u,v);

            //Coordonnee de mon point en ces parametres (u,v)
            surf[index_surface](ku,kv)=vec3(x,y,z);

            // ******************************** //
            // Evaluation de la normale         //
            // ******************************** //

            //Calcul de la derivee en u, puis en v
            vec3 du=vec3(spline_x.diff_u(u,v),spline_y.diff_u(u,v),spline_z.diff_u(u,v));
            vec3 dv=vec3(spline_x.diff_v(u,v),spline_y.diff_v(u,v),spline_z.diff_v(u,v));

            //Calcul de la normale
            vec3 n=du.cross(dv);
            float L=n.norm();
            float epsilon=1e-6;
            if(L>epsilon)
                n/=L;
            else
                n=vec3(1,0,0);

            //Valeur de la normale en ce point
            surf[index_surface].set_normal(ku,kv,n);
        }
    }

}





//************************************************************//
//*** Calcul des courbes splines sur la surface (sur les frontieres des patchs
//************************************************************//
void scene::compute_curve_spline()
{
    //N samples
    unsigned int N=50;

    const unsigned int& Nu=g0.size_u();
    const unsigned int& Nv=g0.size_v();

    unsigned int N_patch_u=Nu-3;
    unsigned int N_patch_v=Nv-3;


    if(curve_spline.size()!=N_patch_u+N_patch_v+2)
        curve_spline.resize(N_patch_u+N_patch_v+2);

    //courbes le long des directions v
    for(unsigned int ku=0;ku<N_patch_u;++ku)
    {

        if(curve_spline[ku].size()!=N*N_patch_v)
            curve_spline[ku].resize(N*N_patch_v);
        if(curve_spline[N_patch_u].size()!=N*N_patch_v)
            curve_spline[N_patch_u].resize(N*N_patch_v);

        for(unsigned int kv=0;kv<N_patch_v;++kv)
        {
            patch4 p=g0.get_patch(ku,kv);

            //les coordonnees des sommets du patch courant
            matrix4 Mx=p.get_x();
            matrix4 My=p.get_y();
            matrix4 Mz=p.get_z();

            //les evaluateurs
            evaluator_spline spline_x(Mx);
            evaluator_spline spline_y(My);
            evaluator_spline spline_z(Mz);

            //Variation suivant v
            for(unsigned int k=0;k<N;++k)
            {
                double t=double(k)/(N-1);

                //La courbe varie suivant (u=0,v=t)
                double x=spline_x(0,t);
                double y=spline_y(0,t);
                double z=spline_z(0,t);

                curve_spline[ku][k+N*kv]=vec3(x,y,z);

                //cas particulier du dernier bord pour (u=1,v=t)
                if(ku==N_patch_u-1)
                {
                    double x=spline_x(1,t);
                    double y=spline_y(1,t);
                    double z=spline_z(1,t);

                    curve_spline[N_patch_u][k+N*kv]=vec3(x,y,z);
                }
            }
        }
    }

    for(unsigned int kv=0;kv<N_patch_v;++kv)
    {
        if(curve_spline[N_patch_u+kv+1].size()!=N*N_patch_u)
            curve_spline[N_patch_u+kv+1].resize(N*N_patch_u);
        if(curve_spline[N_patch_u+N_patch_v+1].size()!=N*N_patch_u)
            curve_spline[N_patch_u+N_patch_v+1].resize(N*N_patch_u);

        //Variation suivant v
        for(unsigned int ku=0;ku<N_patch_u;++ku)
        {
            patch4 p=g0.get_patch(ku,kv);

            //les coordonnees des sommets du patch courant
            matrix4 Mx=p.get_x();
            matrix4 My=p.get_y();
            matrix4 Mz=p.get_z();

            //les evaluateurs
            evaluator_spline spline_x(Mx);
            evaluator_spline spline_y(My);
            evaluator_spline spline_z(Mz);

            for(unsigned int k=0;k<N;++k)
            {
                double t=double(k)/(N-1);

                //La courbe varie suivant (u=t,v=0)
                double x=spline_x(t,0);
                double y=spline_y(t,0);
                double z=spline_z(t,0);

                curve_spline[N_patch_u+kv+1][k+N*ku]=vec3(x,y,z);

                //cas particulier du dernier bord pour (u=t,v=1)
                if(kv==N_patch_v-1)
                {
                    double x=spline_x(t,1);
                    double y=spline_y(t,1);
                    double z=spline_z(t,1);

                    curve_spline[N_patch_u+N_patch_v+1][k+N*ku]=vec3(x,y,z);
                }
            }
        }
    }

}


//************************************************************************//
//*** Ici, on affiche ce que l'on veut.                                ***//
//*** Cette fonction est appellee un grand nombre de fois par secondes ***//
//************************************************************************//
void scene::draw_scene()
{


    glEnable(GL_LIGHTING);//commande OpenGL pour l'eclairage



    //update automatique apres un certain nombre d'affichage (utile lors de l'ajout de lignes/colonnes de la grille)
    if(pick.get_is_up_to_date()==false)
    {
        ++counter_update;
        if(counter_update>600)
        {
            update_high_res();
            counter_update=0;
        }
    }



    //Affichage de la surface spline
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(6,6);
    for(unsigned int k=0;k<surf.size();++k)
    {
        //affichage des patchs en couleur
        if(gui_visible.get_colorpatch_visible()==true)
        {
            unsigned int N_color_table=color_table.size();
            const vec3& c=color_table[k%N_color_table];
            glColor3fv(c.pointer());
        }
        else
            glColor3d(0.5,0.9,0.8);

        //affiche de la surface proprement dit
        if(gui_visible.get_surface_visible()==true)
            opengl_drawer::draw(surf[k]);
    }
    glDisable(GL_POLYGON_OFFSET_FILL);

    //affichage des courbes de delimitation
    glDisable(GL_LIGHTING);
    glColor3d(0,0,0);
    glLineWidth(2.0);
    if(gui_visible.get_curves_visible()==true)
        opengl_drawer::draw_curve(curve_spline);


    //affichage de la grille de controle
    glLineWidth(1.5);
    glDisable(GL_LIGHTING);
    glColor3d(0.5,0.5,0.8);
    if(gui_visible.get_grid_visible()==true)
        opengl_drawer::draw(g0);
    glEnable(GL_LIGHTING);


    //affichage des positions de la grille en tant que sphere
    glEnable(GL_LIGHTING);
    if(gui_visible.get_gridposition_visible()==true)
    {
        const unsigned int& Nu=g0.size_u();
        const unsigned int& Nv=g0.size_v();
        for(unsigned int ku=0;ku<Nu;++ku)
        {
            for(unsigned int kv=0;kv<Nv;++kv)
            {
                const vec3& t=g0(ku,kv);

                //translation rapide en OpenGL par Push/Pop-Matrix
                glPushMatrix();
                glTranslated(t.x(),t.y(),t.z());
                glColor3d(1,0,0);

                //modification de la couleur lors du picking d'une sphere
                if(pick.get_is_picked()==true)
                {
                    const std::map<unsigned int,std::pair<unsigned int,unsigned int> >& picked_index=pick.get_picked_index();
                    if(picked_index.find(ku+Nu*kv)!=picked_index.end())
                        glColor3d(1,1,0);
                }

                //affichage de la sphere
                opengl_drawer::draw(sphere);
                glPopMatrix();
            }
        }
    }
    glDisable(GL_LIGHTING);




}








// Methodes de selection par picking des sommets de la grille
void scene::picking_grid(const vec3& center_ray,const vec3& direction_ray)
{
    const unsigned int& Nu=g0.size_u();
    const unsigned int& Nv=g0.size_v();

    double t_inter=+9999;
    bool is_inter=false;
    unsigned int index_inter_u=0;
    unsigned int index_inter_v=0;

    for(unsigned int ku=0;ku<Nu;++ku)
    {
        for(unsigned int kv=0;kv<Nv;++kv)
        {
            const vec3& x0=g0(ku,kv);

            //calcul des interesection par lance de rayons
            std::vector<double> inter=intersection_helper::sphere_ray(x0,0.08,center_ray,direction_ray);
            for(unsigned int kt=0,Nt=inter.size();kt<Nt;++kt)
            {
                const double& t=inter[kt];
                if(t>=0 && t<t_inter)
                {
                    t_inter=t;
                    index_inter_u=ku;
                    index_inter_v=kv;
                    is_inter=true;
                }
            }

        }
    }

    //si intersection: ajout dans la liste des sommets en intersection
    if(is_inter==true)
    {
        pick.set_is_picked(true);
        std::map<unsigned int,std::pair<unsigned int,unsigned int> > map_index=pick.get_picked_index();
        unsigned int k_index=index_inter_u+Nu*index_inter_v;

        if(map_index.find(k_index)==map_index.end())
        {
            map_index.insert(std::make_pair(k_index,std::make_pair(index_inter_u,index_inter_v)));
            pick.set_picked_index(map_index);
        }
    }
    else
    {
        pick.set_is_picked(false);
        pick.set_picked_index(std::map<unsigned int,std::pair<unsigned int,unsigned int> >());
    }


}

//translation des points selectionnes de la grille
void scene::translate_picking(const cpe::vec3 t)
{
    if(pick.get_is_picked()==true)
    {

        const std::map<unsigned int,std::pair<unsigned int,unsigned int> >& map_index=pick.get_picked_index();
        for(std::map<unsigned int,std::pair<unsigned int,unsigned int> >::const_iterator it=map_index.begin();it!=map_index.end();++it)
        {
            const unsigned int& ku=it->second.first;
            const unsigned int& kv=it->second.second;

            g0(ku,kv)+=t;
        }

        build_spline_surface(N_sample_low_res);
        for(unsigned int k=0;k<surf.size();++k)
            surf[k].update_connectivity();
        compute_curve_spline();

        pick.set_is_up_to_date(false);
    }
}

//Calcul complet de la surface a haute resolution
void scene::update_high_res()
{
    if(pick.get_is_up_to_date()==false)
    {
        build_spline_surface(N_sample_high_res);
        for(unsigned int k=0;k<surf.size();++k)
            surf[k].update_connectivity();
        compute_curve_spline();


        pick.set_is_up_to_date(true);

    }

}

//Calcul complet de la surface a basse resolution
void scene::update_low_res()
{
    build_spline_surface(N_sample_low_res);
    for(unsigned int k=0;k<surf.size();++k)
        surf[k].update_connectivity();
    compute_curve_spline();
    pick.set_is_up_to_date(false);
    counter_update=0;
}

//initialisation de la classe scene
scene::scene()
{

    N_sample_high_res=20;
    N_sample_low_res=6;
    counter_update=0;

    //nom de la grille
    filename_grid="data/ma_grille.txt";

    init_gui_visibility();
    init_color_table();

}

//ajout d'une ligne dans la grille
void scene::add_u_back_line()
{
    g0.add_back_u();
    update_low_res();
}
//ajout d'une ligne dans la grille
void scene::add_u_front_line()
{
    g0.add_front_u();
    update_low_res();
}
//ajout d'une ligne dans la grille
void scene::add_v_back_line()
{
    g0.add_back_v();
    update_low_res();
}
//ajout d'une ligne dans la grille
void scene::add_v_front_line()
{
    g0.add_front_v();
    update_low_res();
}


//supression d'une ligne dans la grille
void scene::erase_u_back_line()
{
    if(g0.size_u()>4)
        g0.suppress_back_u();
    else
        std::cout<<"Cannot erase u line"<<std::endl;
    update_low_res();
}
//supression d'une ligne dans la grille
void scene::erase_u_front_line()
{
    if(g0.size_u()>4)
        g0.suppress_front_u();
    else
        std::cout<<"Cannot erase u line"<<std::endl;
    update_low_res();
}
//supression d'une ligne dans la grille
void scene::erase_v_back_line()
{
    if(g0.size_v()>4)
        g0.suppress_back_v();
    else
        std::cout<<"Cannot erase v line"<<std::endl;
    update_low_res();
}
//supression d'une ligne dans la grille
void scene::erase_v_front_line()
{
    if(g0.size_v()>4)
        g0.suppress_front_v();
    else
        std::cout<<"Cannot erase v line"<<std::endl;
    update_low_res();
}
//duplication des sommets de la scene
void scene::duplicate_boundary()
{
    g0.duplicate_boundary();
    update_low_res();
}
//chargement d'une grille à partir d'un fichier
void scene::load_grid()
{
    //verifie que l'on puisse lire le fichier designe
    std::ifstream fstream(filename_grid.c_str());
    if(fstream.good()==false)
    {
        std::cout<<"Error cannot load grid at location: "<<filename_grid<<std::endl;
        return ;
    }
    fstream.close();


    g0=grid::read(filename_grid);
    std::cout<<"Load grid at location ["<<filename_grid<<"]"<<std::endl;

    pick.set_is_up_to_date(false);
    update_high_res();

}

//sauvegarde d'une grille à partir d'un fichier
void scene::save_grid() const
{
    //verifie que l'on puisse écrire sur le fichier designe
    std::ofstream fstream(filename_grid.c_str());
    if(fstream.good()==false)
    {
        std::cout<<"Error cannot save grid at location: "<<filename_grid<<std::endl;
        return ;
    }
    fstream.close();


    g0.write(filename_grid);
    std::cout<<"Saved grid at location ["<<filename_grid<<"]"<<std::endl;
}


//retourne l'interface de sauvegarde des parametres de la gui
gui_interface& scene::gui_visible_interface()
{
    return gui_visible;
}
//initialisation des couleurs de la table
void scene::init_color_table()
{
    color_table.push_back(vec3(0.5,0.8,0.4));
    color_table.push_back(vec3(0.7,0.2,0.1));
    color_table.push_back(vec3(0.4,0.9,0.1));
    color_table.push_back(vec3(0.2,0.2,0.8));
    color_table.push_back(vec3(0.5,0.5,  0));
    color_table.push_back(vec3(0.5,  0,0.5));
    color_table.push_back(vec3(0  ,0.5,0.5));
    color_table.push_back(vec3(0.1,0.4,0.9));
    color_table.push_back(vec3(0.5,0.5,0.5));
}
//initialisation des parametres de visibilite de la scene
void scene::init_gui_visibility()
{
    gui_visible.set_grid_visible(true);
    gui_visible.set_gridposition_visible(true);
    gui_visible.set_curves_visible(true);
    gui_visible.set_colorpatch_visible(false);
    gui_visible.set_surface_visible(true);
}
//prechargement de la sphere pour les sommets de la grille
void scene::preload_sphere()
{
    std::string filename="data/sphere.off";
    std::ifstream stream(filename.c_str());
    if(stream.is_open()==false || stream.good()==false)
    {
        std::cout<<"Fichier "<<filename<<" non accessible, verifiez le chemin d'execution"<<std::endl;
        abort();
    }
    stream.close();

    sphere.load_file("data/sphere.off");
    sphere.scale(0.06,0.06,0.06);
    sphere.compute_normal();
}


